﻿using System;
using System.Collections.Generic;
using System.Threading;
using DHI.Solutions.Generic;
using MyNoteManager.Interfaces;

namespace MyNoteManager.Business
{
    public class MyNoteModule : IMyNoteModule
    {
        private readonly bool _xThreadSafe;
        private object _moduleSyncObj = new object();

        private IApplication _app;
        private IList<IEntityProvider> _entityProviders;

        private IMyNoteList _myNoteList;
        private IMyNoteGroupList _myNoteGroupList;

        #region constructor
        /// <summary>
        /// Initializes a new instance of the Module class.
        /// </summary>
        public MyNoteModule()
        {
            _app = null;
            _xThreadSafe = DSSAppSetting.GetBoolValue("DHI.Solutions.Core", "XThreadSafe", false);
        }
        #endregion

        #region public properties and methods
        /// <inheritdoc />
        public string Name
        {
            get { return Resources.Module_Name; }
        }

        /// <inheritdoc />
        public string Description
        {
            get { return Resources.Module_Description; }
        }

        /// <inheritdoc />
        public bool Enabled
        {
            get { return true; }
        }

        /// <inheritdoc />
        public bool IsSystemModule
        {
            get { return false; }
        }

        /// <inheritdoc />
        public void ShutDown()
        {
            _OnModuleShutDown();
        }

        /// <inheritdoc />
        public void StartUp(DHI.Solutions.Generic.IApplication app)
        {
            _app = app;
            _OnModuleStarted();
        }
        #endregion

        #region lists
        /// <inheritdoc />
        public IMyNoteList MyNoteList
        {
            get
            {
                if (_myNoteList == null)
                {
                    _LockModule();
                    try
                    {
                        if (_myNoteList == null)
                        {
                            _myNoteList = (IMyNoteList)CreateNewList<IMyNote>();
                        }
                    }
                    finally
                    {
                        _UnlockModule();
                    }
                }

                return _myNoteList;
            }
        }

        public IMyNoteGroupList MyNoteGroupList
        {
            get
            {
                if (_myNoteGroupList == null)
                {
                    _LockModule();
                    try
                    {
                        if (_myNoteGroupList == null)
                        {
                            _myNoteGroupList = (IMyNoteGroupList)CreateNewList<IMyNoteGroup>();
                        }
                    }
                    finally
                    {
                        _UnlockModule();
                    }
                }

                return _myNoteGroupList;
            }
        }
        #endregion

        #region IListModule implementation

        /// <inheritdoc />
        public IList<IEntityProvider> EntityProviders
        {
            get
            {
                if (_entityProviders == null)
                {
                    _LockModule();
                    try
                    {
                        if (_entityProviders == null)
                        {
                            _entityProviders = new List<IEntityProvider>() { MyNoteList, MyNoteGroupList };
                        }
                    }
                    finally
                    {
                        _UnlockModule();
                    }
                }

                return _entityProviders;
            }
        }

        /// <summary>
        /// Creates a new list for a given type of entity.
        /// </summary>
        /// <typeparam name="T">An entity type supported by the module (IIndicator, IIndicatorGroup).</typeparam>
        /// <returns>The reference to a newly created list.</returns>
        public IEntityList<T> CreateNewList<T>() where T : IEntity
        {
            if (typeof(T) == typeof(IMyNoteList))
            {
                return (IEntityList<T>)new MyNoteList(_app);
            }
            else if (typeof(T) == typeof(IMyNote))
            {
                return (IEntityList<T>)new MyNoteList(_app);
            }
            else if (typeof(T) == typeof(IMyNoteGroup))
            {
                return (IEntityList<T>)new MyNoteGroupList(_app, this);
            }
            else
            {
                throw new ArgumentException(string.Format("Entity type {0} is not supported by MyNote Module", typeof(T).Name));
            }
        }
        #endregion

        #region event handlers
        public event EventHandler<DHI.Solutions.Generic.ModuleEventArgs> ModuleShutDown;

        public event EventHandler<DHI.Solutions.Generic.ModuleEventArgs> ModuleStarted;
        #endregion

        #region private
        private void _OnModuleStarted()
        {
            // Raise the ModuleStarted Event
            if (ModuleStarted != null)
            {
                ModuleStarted(this, new ModuleEventArgs(this));
            }
        }

        private void _OnModuleShutDown()
        {
            // Raise the ModuleShudDown Event
            if (ModuleShutDown != null)
            {
                ModuleShutDown(this, new ModuleEventArgs(this));
            }
        }

        private void _LockModule()
        {
            if (_xThreadSafe)
            {
                Monitor.Enter(_moduleSyncObj);
            }
        }

        private void _UnlockModule()
        {
            if (_xThreadSafe)
            {
                Monitor.Exit(_moduleSyncObj);
            }
        }
        #endregion
    }
}